;;; lib.ss
;;; Copyright (c) 2005, R. Kent Dybvig, Simon L. Peyton Jones, and Amr Sabry

;;; Definitions of some common delimited control operators

(define P (newPrompt))

(define-syntax prompt
  (syntax-rules ()
    [(_ e1 e2 ...)
     (pushPrompt P e1 e2 ...)]))

(define F ; aka +F-
 ; captures up to but not including prompt
 ; aborts up to but not including prompt
 ; i.e., leaves prompt in uncaptured/unaborted portion of the continuation
  (lambda (proc)
    (withSubCont P
      (lambda (s)
        (pushPrompt P
          (proc (lambda (v) (pushSubCont s v))))))))

(define-syntax control
  (syntax-rules ()
    ((_ x e1 e2 ...) (F (lambda (x) e1 e2 ...)))))

;;; variations on F
;;; -F-: aborts past prompt; does not reinstate prompt on throw
;;; -F+: aborts past prompt; does reinstate prompt on throw
;;; +F-: (original F) leaves prompt; does not reinstate prompt on throw
;;; +F+: (shift) leaves prompt; reinstates on throw

(define subcontinuation
  (lambda (p restore? s)
     (lambda (v)
       (if restore?
           (pushPrompt p (pushSubCont s v))
           (pushSubCont s v)))))

(define controller
  (lambda (p leave? restore?)
    (lambda (proc)
      (withSubCont p
        (lambda (s)
          (if leave?
              (pushPrompt p (proc (subcontinuation p restore? s)))
              (proc (subcontinuation p restore? s))))))))

(define -F- (controller P #f #f))
(define -F+ (controller P #f #t))
(define +F- (controller P #t #f))
;(define F +F-)
(define +F+ (controller P #t #t))

; alternatively, once we have -F-
;(define -F+
;  (lambda (p)
;    (-F- (lambda (k)
;           (p (lambda (v)
;                (prompt (k v))))))))
;(define +F-
;  (lambda (p)
;    (-F- (lambda (k)
;           (prompt (p k))))))
;(define +F+
;  (lambda (p)
;    (-F- (lambda (k)
;           (prompt
;             (p (lambda (v)
;                  (prompt (k v)))))))))

;;; shift and reset
(define-syntax reset
  (syntax-rules ()
    ((_ e1 e2 ...) (prompt e1 e2 ...))))

(define-syntax shift
  (syntax-rules ()
    ((_ k e1 e2 ...) (+F+ (lambda (k) e1 e2 ...)))))

;;; spawn
(define spawn
  (lambda (proc)
    (let ((p (newPrompt)))
      (pushPrompt p
        (proc (controller p #f #t))))))
