;;; ex.ss
;;; Copyright (c) 2005, R. Kent Dybvig, Simon L. Peyton Jones, and Amr Sabry

;;; Some simple examples of the operators

#|

Try:

  % scheme basic.ss
  Chez Scheme Version 7.0a
  Copyright (c) 1985-2005 Cadence Research Systems

  > (go)
  >> (load "lib.ss")

At the next >> prompt, copy and paste examples from below, e.g.:

  >> (spawn (lambda (c) (* 2 (c (lambda (f) (* 5 (f 3)))))))
  30

Try also with tail.ss, cps-basic.ss, and cps-tail.ss.

|#

;;; From Section 2.1

((lambda (p)
   (+ 2
      (pushPrompt p
        (if (withSubCont p
              (lambda (k)
                (+ (pushSubCont k #f)
                   (pushSubCont k #t))))
            3
            4))))
 (newPrompt)) ;=> 9

;;; From Section 5.2

(define (tailtest)
  (let ([p (newPrompt)])
    (pushPrompt p
      (withSubCont p
        (lambda (s)
          (pushSubCont s (tailtest)))))))

(tailtest) ;=> <nonterminating>

;; additional tail tests

(define (tailtest1a)
  ; same as above, but push same prompt each time
  (let ([p (newPrompt)])
    (let loop ()
      (pushPrompt p
        (withSubCont p
          (lambda (s)
            (pushSubCont s (loop))))))))

(tailtest1a) ;=> <nonterminating>

(define (tailtest2)
  ; repeatedly capture but don't push an nonempty subcontination
  (let ([p (newPrompt)])
    (let loop ()
      (pushPrompt p
        ((withSubCont p
           (lambda (s)
             (loop))))))))

(tailtest2) ;=> <nonterminating>

(define (tailtest3)
  ; test repeated push of empty subcontinuation
  (let ([p (newPrompt)])
    (pushPrompt p
      (withSubCont p
        (lambda (s)
          (let loop ()
            (pushSubCont s (loop))))))))

(tailtest3) ;=> <nonterminating>

(define (tailtest4)
  ; make sure we can do some other control operations and return
  ; from them in between tail calls
  (let loop ()
    (let ([p (newPrompt)])
      (pushPrompt p
        (pushPrompt p 'prompt)
        (withSubCont p
          (lambda (s)
            (pushSubCont s 'subcont)
            (pushSubCont s (loop))))))))

(tailtest4) ;=> <nonterminating>

;;; shift and reset (Danvy and Filinski 1990)
;;; requires lib.ss

(let ()
  (define emit
    (lambda (n)
      (shift c (cons n (c '*)))))
  (reset (emit 1) (emit 2) (emit 3) '())) ;=> (1 2 3)

;;; spawn (Hieb and Dybvig 1990)
;;; requires lib.ss

(spawn (lambda (c) (* 2 (c (lambda (f) (* 5 (f 3))))))) ;=> 30

(spawn (lambda (c) (* 2 (c (lambda (f) (f (f 3))))))) ;=> 12

((spawn (lambda (c) (* 2 (c (lambda (f) f))))) 3) ;=> 6

(spawn (lambda (c1)
         (* 5 ((spawn (lambda (c2)
                        (* 2 (c2 (lambda (f2)
                                   (lambda ()
                                     (* 3 (c1 (lambda (f3)
                                                7))))))))))))) ;=> 7

(spawn (lambda (c)
         (* 5 ((lambda (x) (c (lambda (f1) x)))
               (* 3 (c (lambda (f2) (* 2 (f2 7))))))))) ;=> 42

((spawn (lambda (c1)
          (* 5 (spawn (lambda (c2)
                        (* 2 ((c1 (lambda (f1)
                                    (lambda (g) (g f1 c2)))))))))))
 (lambda (f1 c2) (f1 (lambda () (* 7 (c2 (lambda (f2)
                                           (* 11 (f2 13))))))))) ;=> 10010

;; test tail recursion
(let f ()
  (spawn (lambda (c) (c (lambda (k) (f)))))) ;=> <nonterminating>

;; test tail recursion
(spawn
  (rec f
    (lambda (c)
      ((c (lambda (k)
            (k (lambda () (f c))))))))) ;=> <nonterminating>

;; exercise "no prompt found" error
(spawn (lambda (c1)
         (* 5 (spawn (lambda (c2)
                       (* 2 (c2 (lambda (f2)
                                  (* 3 (c2 (lambda (f3)
                                             7))))))))))) ;=> <error>

;; exercise "no prompt found" error
(spawn (lambda (c1)
         (* 5 ((spawn (lambda (c2)
                        (* 2 (c2 (lambda (f2)
                                   (lambda ()
                                     (* 3 (c2 (lambda (f3)
                                                7))))))))))))) ;=> <error>
